from typing import ClassVar, Sequence, NamedTuple, overload
from itertools import groupby
from pathlib import Path

from dlt.common.storages.file_storage import FileStorage
from dlt.common.configuration import with_config
from dlt.common.configuration.specs import NormalizeVolumeConfiguration
from dlt.common.storages.versioned_storage import VersionedStorage
from dlt.common.configuration.accessors import config


class TParsedNormalizeFileName(NamedTuple):
    schema_name: str
    table_name: str
    file_id: str


class NormalizeStorage(VersionedStorage):

    STORAGE_VERSION: ClassVar[str] = "1.0.0"
    EXTRACTED_FOLDER: ClassVar[str] = "extracted"  # folder within the volume where extracted files to be normalized are stored

    @with_config(spec=NormalizeVolumeConfiguration, namespaces=("normalize",))
    def __init__(self, is_owner: bool, config: NormalizeVolumeConfiguration = config.value) -> None:
        super().__init__(NormalizeStorage.STORAGE_VERSION, is_owner, FileStorage(config.normalize_volume_path, "t", makedirs=is_owner))
        self.config = config
        if is_owner:
            self.initialize_storage()

    def initialize_storage(self) -> None:
        self.storage.create_folder(NormalizeStorage.EXTRACTED_FOLDER, exists_ok=True)

    def list_files_to_normalize_sorted(self) -> Sequence[str]:
        return sorted(self.storage.list_folder_files(NormalizeStorage.EXTRACTED_FOLDER))

    def group_by_schema(self, files: Sequence[str]) -> "groupby[str, str]":
        return groupby(files, NormalizeStorage.get_schema_name)

    @staticmethod
    def get_schema_name(file_name: str) -> str:
        return NormalizeStorage.parse_normalize_file_name(file_name).schema_name

    @staticmethod
    def build_extracted_file_stem(schema_name: str, table_name: str, file_id: str) -> str:
        # builds file name with the extracted data to be passed to normalize
        return f"{schema_name}.{table_name}.{file_id}"

    @staticmethod
    def parse_normalize_file_name(file_name: str) -> TParsedNormalizeFileName:
        # parse extracted file name and returns (events found, load id, schema_name)
        if not file_name.endswith("jsonl"):
            raise ValueError(file_name)

        parts = Path(file_name).stem.split(".")
        if len(parts) != 3:
            raise ValueError(file_name)
        return TParsedNormalizeFileName(*parts)



from typing import Literal, Optional

from dlt.common.configuration.specs.base_configuration import BaseConfiguration, configspec

TPoolType = Literal["process", "thread", "none"]


@configspec
class PoolRunnerConfiguration(BaseConfiguration):
    pool_type: TPoolType = None  # type of pool to run, must be set in derived configs
    workers: Optional[int] = None  # how many threads/processes in the pool
    run_sleep: float = 0.5  # how long to sleep between runs with workload, seconds
    run_sleep_idle: float = 1.0  # how long to sleep when no more items are pending, seconds
    run_sleep_when_failed: float = 1.0  # how long to sleep between the runs when failed
    is_single_run: bool = False  # should run only once until all pending data is processed, and exit
    wait_runs: int = 0  # how many runs to wait for first data coming in is IS_SINGLE_RUN is set
    exit_on_exception: bool = False  # should exit on exception
    stop_after_runs: int = 10000  # will stop runner with exit code -2 after so many runs, that prevents memory fragmentation



from typing import List, Optional, Tuple, TYPE_CHECKING

from dlt.common.configuration.specs.base_configuration import ContainerInjectableContext, configspec


@configspec(init=True)
class ConfigNamespacesContext(ContainerInjectableContext):
    pipeline_name: Optional[str]
    namespaces: Tuple[str, ...] = ()

    if TYPE_CHECKING:
        # provide __init__ signature when type checking
        def __init__(self, pipeline_name:str = None, namespaces: Tuple[str, ...] = ()) -> None:
            ...



import inspect
import types
import asyncio
import makefun
from asyncio import Future
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from threading import Thread
from typing import Any, Optional, Sequence, Union, Callable, Iterable, Iterator, List, NamedTuple, Awaitable, Tuple, Type, TYPE_CHECKING

from dlt.common.configuration import configspec
from dlt.common.configuration.inject import with_config
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.typing import AnyFun, TDataItems
from dlt.common.utils import get_callable_name

from dlt.extract.exceptions import CreatePipeException, InvalidStepFunctionArguments, ParametrizedResourceUnbound, PipeItemProcessingError, PipeNotBoundToData
from dlt.extract.typing import DataItemWithMeta, FilterItemFunction, FilterItemFunctionWithMeta, TPipedDataItems


if TYPE_CHECKING:
    TItemFuture = Future[Union[TDataItems, DataItemWithMeta]]
else:
    TItemFuture = Future

from dlt.common.time import sleep


class PipeItem(NamedTuple):
    item: TDataItems
    step: int
    pipe: "Pipe"
    meta: Any


class ResolvablePipeItem(NamedTuple):
    # mypy unable to handle recursive types, ResolvablePipeItem should take itself in "item"
    item: Union[TPipedDataItems, Iterator[TPipedDataItems]]
    step: int
    pipe: "Pipe"
    meta: Any


class FuturePipeItem(NamedTuple):
    item: TItemFuture
    step: int
    pipe: "Pipe"
    meta: Any


class SourcePipeItem(NamedTuple):
    item: Union[Iterator[TPipedDataItems], Iterator[ResolvablePipeItem]]
    step: int
    pipe: "Pipe"
    meta: Any


# pipeline step may be iterator of data items or mapping function that returns data item or another iterator
TPipeStep = Union[
    Iterable[TPipedDataItems],
    Iterator[TPipedDataItems],
    Callable[[TDataItems, Optional[Any]], TPipedDataItems],
    Callable[[TDataItems, Optional[Any]], Iterator[TPipedDataItems]],
    Callable[[TDataItems, Optional[Any]], Iterator[ResolvablePipeItem]]
]


# def _meta_wrapper(f: AnyFun, *args: Any, **kwargs: Any) -> Any:
#     """Wraps a single argument transformation expecting item in two argument transformation also accepting meta"""
#     # TODO: must find first arg or kwarg and pass it to f
#     return f(args[0])


class ForkPipe:
    def __init__(self, pipe: "Pipe", step: int = -1) -> None:
        self._pipes: List[Tuple["Pipe", int]] = []
        self.add_pipe(pipe, step)

    def add_pipe(self, pipe: "Pipe", step: int = -1) -> None:
        if pipe not in self._pipes:
            self._pipes.append((pipe, step))

    def has_pipe(self, pipe: "Pipe") -> bool:
        return pipe in [p[0] for p in self._pipes]

    def __call__(self, item: TDataItems, meta: Any) -> Iterator[ResolvablePipeItem]:
        for i, (pipe, step) in enumerate(self._pipes):
            _it = item if i == 0 else deepcopy(item)
            # always start at the beginning
            yield ResolvablePipeItem(_it, step, pipe, meta)


class FilterItem:
    _filter_f_meta: FilterItemFunctionWithMeta = None
    _filter_f: FilterItemFunction = None

    def __init__(self, filter_f: Union[FilterItemFunctionWithMeta, FilterItemFunction]) -> None:
        # TODO: extract this into a helper function
        # inspect the signature
        sig = inspect.signature(filter_f)
        # TODO: use TypeGuard here to get rid of type ignore
        if len(sig.parameters) == 1:
            self._filter_f = filter_f  # type: ignore
        else:  # TODO: do better check
            self._filter_f_meta = filter_f  # type: ignore

    def __call__(self, item: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        # item may be a list TDataItem or a single TDataItem
        if isinstance(item, list):
            if self._filter_f_meta:
                item = [i for i in item if self._filter_f_meta(i, meta)]
            else:
                item = [i for i in item if self._filter_f(i)]
            if not item:
                # item was fully consumed by the filter
                return None
            return item
        else:
            if self._filter_f_meta:
                return item if self._filter_f_meta(item, meta) else None
            else:
                return item if self._filter_f(item) else None


class Pipe:
    def __init__(self, name: str, steps: List[TPipeStep] = None, parent: "Pipe" = None) -> None:
        self.name = name
        self._steps: List[TPipeStep] = []
        self._pipe_id = f"{name}_{id(self)}"
        self.parent = parent
        # add the steps, this will check and mod transformations
        if steps:
            for step in steps:
                self.add_step(step)

    @classmethod
    def from_data(cls, name: str, gen: Union[Iterable[TPipedDataItems], Iterator[TPipedDataItems], AnyFun], parent: "Pipe" = None) -> "Pipe":
        return cls(name, [gen], parent=parent)

    @property
    def is_empty(self) -> bool:
        """Checks if pipe contains any steps"""
        return len(self._steps) == 0

    @property
    def has_parent(self) -> bool:
        """Checks if pipe is connected to parent pipe from which it takes data items. Connected pipes are created from transformer resources"""
        return self.parent is not None

    @property
    def is_data_bound(self) -> bool:
        """Checks if pipe is bound to data and can be iterated. Pipe is bound if has a parent that is bound xor is not empty."""
        if self.has_parent:
            return self.parent.is_data_bound
        else:
            return not self.is_empty

    @property
    def head(self) -> TPipeStep:
        return self._steps[0]

    @property
    def tail(self) -> TPipeStep:
        return self._steps[-1]

    @property
    def steps(self) -> List[TPipeStep]:
        return self._steps

    def __getitem__(self, i: int) -> TPipeStep:
        return self._steps[i]

    def __len__(self) -> int:
        return len(self._steps)

    def fork(self, child_pipe: "Pipe", child_step: int = -1) -> "Pipe":
        if len(self._steps) == 0:
            raise CreatePipeException(self.name, f"Cannot fork to empty pipe {child_pipe}")
        fork_step = self.tail
        if not isinstance(fork_step, ForkPipe):
            fork_step = ForkPipe(child_pipe, child_step)
            self.add_step(fork_step)
        else:
            if not fork_step.has_pipe(child_pipe):
                fork_step.add_pipe(child_pipe, child_step)
        return self

    def clone(self) -> "Pipe":
        p = Pipe(self.name, [], self.parent)
        p._steps = self._steps.copy()
        # clone shares the id with the original
        p._pipe_id = self._pipe_id
        return p

    # def backup(self) -> None:
    #     if self.has_backup:
    #         raise PipeBackupException("Pipe backup already exists, restore pipe first")
    #     self._backup_steps = self._steps.copy()

    # @property
    # def has_backup(self) -> bool:
    #     return self._backup_steps is not None


    # def restore(self) -> None:
    #     if not self.has_backup:
    #         raise PipeBackupException("No pipe backup to restore")
    #     self._steps = self._backup_steps
    #     self._backup_steps = None

    def add_step(self, step: TPipeStep) -> "Pipe":
        step_no = len(self._steps)
        if step_no == 0 and not self.has_parent:
            self._add_head(step)
        else:
            # step must be a callable: a transformer or a transformation
            if isinstance(step, (Iterable, Iterator)):
                if self.has_parent:
                    raise CreatePipeException(self.name, "Iterable or Iterator cannot be a step in transformer pipe")
                else:
                    raise CreatePipeException(self.name, "Iterable or Iterator can only be a first step in resource pipe")
            self._add_transformer_step(step_no, step)
        return self

    def full_pipe(self) -> "Pipe":
        # prevent creating full pipe with unbound heads
        self._ensure_step_function(0, self.head)

        if self.has_parent:
            steps = self.parent.full_pipe().steps
        else:
            steps = []

        steps.extend(self._steps)
        p = Pipe(self.name, [])
        # set the steps so they are not evaluated again
        p._steps = steps
        # return pipe with resolved dependencies
        return p

    def evaluate_head(self) -> None:
        """Lazily evaluate head of the pipe when creating PipeIterator. Allows creating multiple use pipes from generator functions and lists"""
        if not self.is_data_bound:
            raise PipeNotBoundToData(self.name, self.has_parent)

        head = self.head
        if not self.has_parent:
            # if pipe head is callable then call it
            if callable(head):
                try:
                    # must be parameterless callable or parameters must have defaults
                    self._steps[0] = head()  # type: ignore
                except TypeError:
                    raise ParametrizedResourceUnbound(self.name, get_callable_name(head), inspect.signature(head))
            elif isinstance(head, Iterable):
                self._steps[0] = iter(head)
        else:
            # verify if transformer can be called
            self._ensure_step_function(0, head)

    def _add_head(self, step: TPipeStep) -> None:
        # first element must be Iterable, Iterator or Callable in resource pipe
        if not isinstance(step, (Iterable, Iterator)) and not callable(step):
            raise CreatePipeException(self.name, "A head of a resource pipe must be Iterable or Iterator or Callable that can be called without arguments")
        self._steps.append(step)

    def _add_transformer_step(self, step_no: int, step: TPipeStep) -> None:
        if not callable(step):
            raise CreatePipeException(self.name, "Pipe step must be a callable taking exactly one data item as argument")
        else:
            # check the signature
            sig = inspect.signature(step)
            sig_arg_count = len(sig.parameters)
            callable_name = get_callable_name(step)
            if sig_arg_count == 0:
                raise InvalidStepFunctionArguments(self.name, callable_name, sig, "Function takes no arguments")
            # see if meta is present in kwargs
            meta_arg = next((p for p in sig.parameters.values() if p.name == "meta"), None)
            if meta_arg is not None:
                if meta_arg.kind not in (meta_arg.KEYWORD_ONLY, meta_arg.POSITIONAL_OR_KEYWORD):
                    raise InvalidStepFunctionArguments(self.name, callable_name, sig, "'meta' cannot be pos only argument '")
            elif step_no > 0 or sig_arg_count == 1:  # transformer always has meta added by the resource
                # add meta parameter, in case of other number of params (0 or >1) let the pipe code fail later
                orig_step = step

                def _partial(*args: Any, **kwargs: Any) -> Any:
                    # orig step does not have meta
                    del kwargs["meta"]
                    return orig_step(*args, **kwargs)

                step = makefun.wraps(
                        step,
                        append_args=inspect.Parameter("meta", inspect._ParameterKind.KEYWORD_ONLY, default=None)
                    )(_partial)

            # verify the step callable
            if step_no > 0:
                self._ensure_step_function(step_no, step)

        self._steps.append(step)

    def _ensure_step_function(self, step_no: int, step: TPipeStep) -> None:
        if not callable(step):
            return
        try:
            # get eventually modified sig
            sig = inspect.signature(step)
            sig.bind("item", meta="meta")
        except TypeError as ty_ex:
            callable_name = get_callable_name(step)
            if step_no == 0:
                # show the sig without first argument
                raise ParametrizedResourceUnbound(self.name, callable_name, sig.replace(parameters=list(sig.parameters.values())[1:]), "transformer")
            if step_no > 0:
                raise InvalidStepFunctionArguments(self.name, callable_name, sig, str(ty_ex))

    def __repr__(self) -> str:
        if self.has_parent:
            bound_str = " data bound to " + repr(self.parent)
        else:
            bound_str = ""
        return f"Pipe {self.name} ({self._pipe_id})[steps: {len(self._steps)}] at {id(self)}{bound_str}"


class PipeIterator(Iterator[PipeItem]):

    @configspec
    class PipeIteratorConfiguration(BaseConfiguration):
        max_parallel_items: int = 100
        workers: int = 5
        futures_poll_interval: float = 0.01

    def __init__(self, max_parallel_items: int, workers: int, futures_poll_interval: float) -> None:
        self.max_parallel_items = max_parallel_items
        self.workers = workers
        self.futures_poll_interval = futures_poll_interval

        self._async_pool: asyncio.AbstractEventLoop = None
        self._async_pool_thread: Thread = None
        self._thread_pool: ThreadPoolExecutor = None
        self._sources: List[SourcePipeItem] = []
        self._futures: List[FuturePipeItem] = []

    @classmethod
    @with_config(spec=PipeIteratorConfiguration)
    def from_pipe(cls, pipe: Pipe, *, max_parallel_items: int = 100, workers: int = 5, futures_poll_interval: float = 0.01) -> "PipeIterator":
        # join all dependent pipes
        if pipe.parent:
            pipe = pipe.full_pipe()
        # clone pipe to allow multiple iterations on pipe based on iterables/callables
        pipe = pipe.clone()
        # head must be iterator
        pipe.evaluate_head()
        assert isinstance(pipe.head, Iterator)
        # create extractor
        extract = cls(max_parallel_items, workers, futures_poll_interval)
        # add as first source
        extract._sources.append(SourcePipeItem(pipe.head, 0, pipe, None))
        return extract

    @classmethod
    @with_config(spec=PipeIteratorConfiguration)
    def from_pipes(cls, pipes: Sequence[Pipe], yield_parents: bool = True, *, max_parallel_items: int = 100, workers: int = 5, futures_poll_interval: float = 0.01) -> "PipeIterator":
        extract = cls(max_parallel_items, workers, futures_poll_interval)
        # TODO: consider removing cloning. pipe are single use and may be iterated only once, here we modify an immediately run
        # clone all pipes before iterating (recursively) as we will fork them and this add steps
        pipes = PipeIterator.clone_pipes(pipes)

        def _fork_pipeline(pipe: Pipe) -> None:
            if pipe.parent:
                # fork the parent pipe
                pipe.evaluate_head()
                pipe.parent.fork(pipe)
                # make the parent yield by sending a clone of item to itself with position at the end
                if yield_parents and pipe.parent in pipes:
                    # fork is last step of the pipe so it will yield
                    pipe.parent.fork(pipe.parent, len(pipe.parent) - 1)
                _fork_pipeline(pipe.parent)
            else:
                # head of independent pipe must be iterator
                pipe.evaluate_head()
                assert isinstance(pipe.head, Iterator)
                # add every head as source only once
                if not any(i.pipe == pipe for i in extract._sources):
                    extract._sources.append(SourcePipeItem(pipe.head, 0, pipe, None))

        for pipe in reversed(pipes):
            _fork_pipeline(pipe)

        return extract

    def __next__(self) -> PipeItem:
        pipe_item: Union[ResolvablePipeItem, SourcePipeItem] = None
        # __next__ should call itself to remove the `while` loop and continue clauses but that may lead to stack overflows: there's no tail recursion opt in python
        # https://stackoverflow.com/questions/13591970/does-python-optimize-tail-recursion (see Y combinator on how it could be emulated)
        while True:
            # do we need new item?
            if pipe_item is None:
                # process element from the futures
                if len(self._futures) > 0:
                    pipe_item = self._resolve_futures()
                # if none then take element from the newest source
                if pipe_item is None:
                    pipe_item = self._get_source_item()

                if pipe_item is None:
                    if len(self._futures) == 0 and len(self._sources) == 0:
                        # no more elements in futures or sources
                        raise StopIteration()
                    else:
                        sleep(self.futures_poll_interval)
                    continue

            item = pipe_item.item
            # if item is iterator, then add it as a new source
            if isinstance(item, Iterator):
                # print(f"adding iterable {item}")
                self._sources.append(SourcePipeItem(item, pipe_item.step, pipe_item.pipe, pipe_item.meta))
                pipe_item = None
                continue

            if isinstance(item, Awaitable) or callable(item):
                # do we have a free slot or one of the slots is done?
                if len(self._futures) < self.max_parallel_items or self._next_future() >= 0:
                    if isinstance(item, Awaitable):
                        future = asyncio.run_coroutine_threadsafe(item, self._ensure_async_pool())
                    elif callable(item):
                        future = self._ensure_thread_pool().submit(item)
                    # print(future)
                    self._futures.append(FuturePipeItem(future, pipe_item.step, pipe_item.pipe, pipe_item.meta))  # type: ignore
                    # pipe item consumed for now, request a new one
                    pipe_item = None
                    continue
                else:
                    # print("maximum futures exceeded, waiting")
                    sleep(self.futures_poll_interval)
                # try same item later
                continue

            # if we are at the end of the pipe then yield element
            if pipe_item.step == len(pipe_item.pipe) - 1:
                # must be resolved
                if isinstance(item, (Iterator, Awaitable)) or callable(item):
                    raise PipeItemProcessingError(
                        pipe_item.pipe.name, f"Pipe item at step {pipe_item.step} was not fully evaluated and is of type {type(pipe_item.item).__name__}. This is internal error or you are yielding something weird from resources ie. functions or awaitables.")
                # mypy not able to figure out that item was resolved
                return pipe_item  # type: ignore

            # advance to next step
            step = pipe_item.pipe[pipe_item.step + 1]
            assert callable(step)
            try:
                next_meta = pipe_item.meta
                next_item = step(item, meta=pipe_item.meta)  # type: ignore
                if isinstance(next_item, DataItemWithMeta):
                    next_meta = next_item.meta
                    next_item = next_item.data
            except TypeError as ty_ex:
                raise InvalidStepFunctionArguments(pipe_item.pipe.name, get_callable_name(step), inspect.signature(step), str(ty_ex))
            # create next pipe item if a value was returned. A None means that item was consumed/filtered out and should not be further processed
            if next_item is not None:
                pipe_item = ResolvablePipeItem(next_item, pipe_item.step + 1, pipe_item.pipe, next_meta)
            else:
                pipe_item = None


    def _ensure_async_pool(self) -> asyncio.AbstractEventLoop:
        # lazily create async pool is separate thread
        if self._async_pool:
            return self._async_pool

        def start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
            asyncio.set_event_loop(loop)
            loop.run_forever()

        self._async_pool = asyncio.new_event_loop()
        self._async_pool_thread = Thread(target=start_background_loop, args=(self._async_pool,), daemon=True)
        self._async_pool_thread.start()

        # start or return async pool
        return self._async_pool

    def _ensure_thread_pool(self) -> ThreadPoolExecutor:
        # lazily start or return thread pool
        if self._thread_pool:
            return self._thread_pool

        self._thread_pool = ThreadPoolExecutor(self.workers)
        return self._thread_pool

    def __enter__(self) -> "PipeIterator":
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: types.TracebackType) -> None:

        def stop_background_loop(loop: asyncio.AbstractEventLoop) -> None:
            loop.stop()

        for f, _, _, _ in self._futures:
            if not f.done():
                f.cancel()
        print("stopping loop")
        if self._async_pool:
            self._async_pool.call_soon_threadsafe(stop_background_loop, self._async_pool)
            print("joining thread")
            self._async_pool_thread.join()
            self._async_pool = None
            self._async_pool_thread = None
        if self._thread_pool:
            self._thread_pool.shutdown(wait=True)
            self._thread_pool = None

    def _next_future(self) -> int:
        return next((i for i, val in enumerate(self._futures) if val.item.done()), -1)

    def _resolve_futures(self) -> ResolvablePipeItem:
        # no futures at all
        if len(self._futures) == 0:
            return None

        # anything done?
        idx = self._next_future()
        if idx == -1:
            # nothing done
            return None

        future, step, pipe, meta = self._futures.pop(idx)

        if future.cancelled():
            # get next future
            return self._resolve_futures()

        if future.exception():
            raise future.exception()

        item = future.result()
        if isinstance(item, DataItemWithMeta):
            return ResolvablePipeItem(item.data, step, pipe, item.meta)
        else:
            return ResolvablePipeItem(item, step, pipe, meta)

    def _get_source_item(self) -> ResolvablePipeItem:
        # no more sources to iterate
        if len(self._sources) == 0:
            return None

        # get items from last added iterator, this makes the overall Pipe as close to FIFO as possible
        gen, step, pipe, meta = self._sources[-1]
        try:
            item = next(gen)
            # full pipe item may be returned, this is used by ForkPipe step
            # to redirect execution of an item to another pipe
            if isinstance(item, ResolvablePipeItem):
                return item
            else:
                # keep the item assigned step and pipe when creating resolvable item
                if isinstance(item, DataItemWithMeta):
                    return ResolvablePipeItem(item.data, step, pipe, item.meta)
                else:
                    return ResolvablePipeItem(item, step, pipe, meta)
        except StopIteration:
            # remove empty iterator and try another source
            self._sources.pop()
            return self._get_source_item()

    @staticmethod
    def clone_pipes(pipes: Sequence[Pipe]) -> Sequence[Pipe]:
        # will clone the pipes including the dependent ones
        cloned_pipes = [p.clone() for p in pipes]
        cloned_pairs = {id(p): c for p, c in zip(pipes, cloned_pipes)}

        for clone in cloned_pipes:
            while True:
                if not clone.parent:
                    break
                # if already a clone
                if clone.parent in cloned_pairs.values():
                    break
                # clone if parent pipe not yet cloned
                if id(clone.parent) not in cloned_pairs:
                    # print("cloning:" + clone.parent.name)
                    cloned_pairs[id(clone.parent)] = clone.parent.clone()
                # replace with clone
                # print(f"replace depends on {clone.name} to {clone.parent.name}")
                clone.parent = cloned_pairs[id(clone.parent)]
                # recurr with clone
                clone = clone.parent

        return cloned_pipes



import inspect
from functools import wraps
from typing import Callable, Dict, Type, Any, Optional, Tuple, TypeVar, overload
from inspect import Signature, Parameter

from dlt.common.typing import DictStrAny, StrAny, TFun, AnyFun
from dlt.common.configuration.resolve import resolve_configuration, inject_namespace
from dlt.common.configuration.specs.base_configuration import BaseConfiguration
from dlt.common.configuration.specs.config_namespace_context import ConfigNamespacesContext
from dlt.common.reflection.spec import spec_from_signature


_LAST_DLT_CONFIG = "_dlt_config"
_ORIGINAL_ARGS = "_dlt_orig_args"
TConfiguration = TypeVar("TConfiguration", bound=BaseConfiguration)
# keep a registry of all the decorated functions
_FUNC_SPECS: Dict[int, Type[BaseConfiguration]] = {}



def get_fun_spec(f: AnyFun) -> Type[BaseConfiguration]:
    return _FUNC_SPECS.get(id(f))


@overload
def with_config(func: TFun, /, spec: Type[BaseConfiguration] = None, auto_namespace: bool = False, only_kw: bool = False, namespaces: Tuple[str, ...] = ()) ->  TFun:
    ...


@overload
def with_config(func: None = ..., /, spec: Type[BaseConfiguration] = None, auto_namespace: bool = False, only_kw: bool = False, namespaces: Tuple[str, ...] = ()) ->  Callable[[TFun], TFun]:
    ...


def with_config(func: Optional[AnyFun] = None, /, spec: Type[BaseConfiguration] = None, auto_namespace: bool = False, only_kw: bool = False, namespaces: Tuple[str, ...] = ()) ->  Callable[[TFun], TFun]:

    namespace_f: Callable[[StrAny], str] = None
    # namespace may be a function from function arguments to namespace
    if callable(namespaces):
        namespace_f = namespaces

    def decorator(f: TFun) -> TFun:
        SPEC: Type[BaseConfiguration] = None
        sig: Signature = inspect.signature(f)
        kwargs_arg = next((p for p in sig.parameters.values() if p.kind == Parameter.VAR_KEYWORD), None)
        spec_arg: Parameter = None
        pipeline_name_arg: Parameter = None
        namespace_context = ConfigNamespacesContext()

        if spec is None:
            SPEC = spec_from_signature(f, sig, only_kw)
        else:
            SPEC = spec

        for p in sig.parameters.values():
            # for all positional parameters that do not have default value, set default
            # if hasattr(SPEC, p.name) and p.default == Parameter.empty:
            #     p._default = None  # type: ignore
            if p.annotation is SPEC:
                # if any argument has type SPEC then us it to take initial value
                spec_arg = p
            if p.name == "pipeline_name" and auto_namespace:
                # if argument has name pipeline_name and auto_namespace is used, use it to generate namespace context
                pipeline_name_arg = p
                pipeline_name_arg_default = None if p.default == Parameter.empty else p.default


        @wraps(f)
        def _wrap(*args: Any, **kwargs: Any) -> Any:
            # bind parameters to signature
            bound_args = sig.bind(*args, **kwargs)
            # for calls containing resolved spec in the kwargs, we do not need to resolve again
            config: BaseConfiguration = None
            if _LAST_DLT_CONFIG in kwargs:
                config = last_config(**kwargs)
            else:
                # if namespace derivation function was provided then call it
                nonlocal namespaces
                if namespace_f:
                    namespaces = (namespace_f(bound_args.arguments), )
                # namespaces may be a string
                if isinstance(namespaces, str):
                    namespaces = (namespaces,)
                # if one of arguments is spec the use it as initial value
                if spec_arg:
                    config = bound_args.arguments.get(spec_arg.name, None)
                # resolve SPEC, also provide namespace_context with pipeline_name
                if pipeline_name_arg:
                    namespace_context.pipeline_name = bound_args.arguments.get(pipeline_name_arg.name, pipeline_name_arg_default)
                with inject_namespace(namespace_context):
                    config = resolve_configuration(config or SPEC(), namespaces=namespaces, explicit_value=bound_args.arguments)
            resolved_params = dict(config)
            bound_args.apply_defaults()
            # overwrite or add resolved params
            for p in sig.parameters.values():
                if p.name in resolved_params:
                    bound_args.arguments[p.name] = resolved_params.pop(p.name)
                if p.annotation is SPEC:
                    bound_args.arguments[p.name] = config
            # pass all other config parameters into kwargs if present
            if kwargs_arg is not None:
                bound_args.arguments[kwargs_arg.name].update(resolved_params)
                bound_args.arguments[kwargs_arg.name][_LAST_DLT_CONFIG] = config
                bound_args.arguments[kwargs_arg.name][_ORIGINAL_ARGS] = (args, kwargs)
            # call the function with resolved config
            return f(*bound_args.args, **bound_args.kwargs)

        # register the spec for a wrapped function
        _FUNC_SPECS[id(_wrap)] = SPEC

        return _wrap  # type: ignore

    # See if we're being called as @with_config or @with_config().
    if func is None:
        # We're called with parens.
        return decorator

    if not callable(func):
        raise ValueError("First parameter to the with_config must be callable ie. by using it as function decorator")

    # We're called as @with_config without parens.
    return decorator(func)


def last_config(**kwargs: Any) -> Any:
    return kwargs[_LAST_DLT_CONFIG]


def get_orig_args(**kwargs: Any) -> Tuple[Tuple[Any], DictStrAny]:
    return kwargs[_ORIGINAL_ARGS]  # type: ignore



import threading

from dlt.common import logger
from dlt.common.logger import init_logging_from_config, init_telemetry
from dlt.common.signals import register_signals
from dlt.common.configuration.specs import RunConfiguration

# signals and telemetry should be initialized only once
_INITIALIZED = False


def initialize_runner(C: RunConfiguration) -> None:
    global _INITIALIZED

    # initialize or re-initialize logging with new settings
    init_logging_from_config(C)

    # initialize only once
    if not _INITIALIZED:
        init_telemetry(C)

        if threading.current_thread() is threading.main_thread():
            register_signals()
        else:
            logger.info("Running in daemon thread, signals not enabled")

        _INITIALIZED = True


from collections.abc import Mapping as C_Mapping
from typing import Any, Dict, ContextManager, List, Optional, Sequence, Tuple, Type, TypeVar

from dlt.common.configuration.providers.provider import ConfigProvider
from dlt.common.typing import AnyType, StrAny, TSecretValue, is_final_type, is_optional_type

from dlt.common.configuration.specs.base_configuration import BaseConfiguration, CredentialsConfiguration, is_secret_hint, extract_inner_hint, is_context_inner_hint, is_base_configuration_inner_hint
from dlt.common.configuration.specs.config_namespace_context import ConfigNamespacesContext
from dlt.common.configuration.container import Container
from dlt.common.configuration.specs.config_providers_context import ConfigProvidersContext
from dlt.common.configuration.utils import log_traces, deserialize_value
from dlt.common.configuration.exceptions import (FinalConfigFieldException, LookupTrace, ConfigFieldMissingException, ConfigurationWrongTypeException, ValueNotSecretException, InvalidNativeValue)

TConfiguration = TypeVar("TConfiguration", bound=BaseConfiguration)


def resolve_configuration(config: TConfiguration, *, namespaces: Tuple[str, ...] = (), explicit_value: Any = None, accept_partial: bool = False) -> TConfiguration:
    if not isinstance(config, BaseConfiguration):
        raise ConfigurationWrongTypeException(type(config))

    # try to get the native representation of the top level configuration using the config namespace as a key
    # allows, for example, to store connection string or service.json in their native form in single env variable or under single vault key
    if config.__namespace__ and explicit_value is None:
        initial_hint = TSecretValue if isinstance(config, CredentialsConfiguration) else AnyType
        explicit_value, traces = _resolve_single_value(config.__namespace__, initial_hint, AnyType, None, namespaces, ())
        if isinstance(explicit_value, C_Mapping):
            # mappings cannot be used as explicit values, we want to enumerate mappings and request the fields' values one by one
            explicit_value = None
        else:
            log_traces(None, config.__namespace__, type(config), explicit_value, None, traces)

    return _resolve_configuration(config, namespaces, (), explicit_value, accept_partial)


def inject_namespace(namespace_context: ConfigNamespacesContext, merge_existing: bool = True) -> ContextManager[ConfigNamespacesContext]:
    """Adds `namespace` context to container, making it injectable. Optionally merges the context already in the container with the one provided

    Args:
        namespace_context (ConfigNamespacesContext): Instance providing a pipeline name and namespace context
        merge_existing (bool, optional): Gets `pipeline_name` and `namespaces` from existing context if they are not provided in `namespace` argument. Defaults to True.

    Yields:
        Iterator[ConfigNamespacesContext]: Context manager with current namespace context
    """
    container = Container()
    existing_context = container[ConfigNamespacesContext]

    if merge_existing:
        namespace_context.pipeline_name = namespace_context.pipeline_name or existing_context.pipeline_name
        namespace_context.namespaces = namespace_context.namespaces or existing_context.namespaces

    return container.injectable_context(namespace_context)


def _resolve_configuration(
        config: TConfiguration,
        explicit_namespaces: Tuple[str, ...],
        embedded_namespaces: Tuple[str, ...],
        explicit_value: Any,
        accept_partial: bool
    ) -> TConfiguration:

    # do not resolve twice
    if config.is_resolved():
        return config

    config.__exception__ = None
    try:
        try:
            # use initial value to resolve the whole configuration. if explicit value is a mapping it will be applied field by field later
            if explicit_value and not isinstance(explicit_value, C_Mapping):
                try:
                    config.parse_native_representation(explicit_value)
                except ValueError as v_err:
                    # provide generic exception
                    raise InvalidNativeValue(type(config), type(explicit_value), embedded_namespaces, v_err)
                except NotImplementedError:
                    pass
                # explicit value was consumed
                explicit_value = None

            # if native representation didn't fully resolve the config, we try to resolve field by field
            if not config.is_resolved():
                _resolve_config_fields(config, explicit_value, explicit_namespaces, embedded_namespaces, accept_partial)

            _call_method_in_mro(config, "on_resolved")
            # full configuration was resolved
            config.__is_resolved__ = True
        except ConfigFieldMissingException as cm_ex:
            # store the ConfigEntryMissingException to have full info on traces of missing fields
            config.__exception__ = cm_ex
            _call_method_in_mro(config, "on_partial")
            # if resolved then do not raise
            if config.is_resolved():
                _call_method_in_mro(config, "on_resolved")
            else:
                if not accept_partial:
                    raise
    except Exception as ex:
        # store the exception that happened in the resolution process
        config.__exception__ = ex
        raise

    return config


def _resolve_config_fields(
        config: BaseConfiguration,
        explicit_values: StrAny,
        explicit_namespaces: Tuple[str, ...],
        embedded_namespaces: Tuple[str, ...],
        accept_partial: bool
    ) -> None:

    fields = config.get_resolvable_fields()
    unresolved_fields: Dict[str, Sequence[LookupTrace]] = {}

    for key, hint in fields.items():
        # get default and explicit values
        default_value = getattr(config, key, None)

        if explicit_values:
            explicit_value = explicit_values.get(key)
        else:
            if is_final_type(hint):
                # for final fields default value is like explicit
                explicit_value = default_value
            else:
                explicit_value = None

        current_value, traces = _resolve_config_field(key, hint, default_value, explicit_value, config, config.__namespace__, explicit_namespaces, embedded_namespaces, accept_partial)
        # check if hint optional
        is_optional = is_optional_type(hint)
        # collect unresolved fields
        if not is_optional and current_value is None:
            unresolved_fields[key] = traces
        # set resolved value in config
        if default_value != current_value:
            if is_final_type(hint):
                raise FinalConfigFieldException(type(config).__name__, key)
            setattr(config, key, current_value)

    if unresolved_fields:
        raise ConfigFieldMissingException(type(config).__name__, unresolved_fields)


def _resolve_config_field(
        key: str,
        hint: Type[Any],
        default_value: Any,
        explicit_value: Any,
        config: BaseConfiguration,
        config_namespace: str,
        explicit_namespaces: Tuple[str, ...],
        embedded_namespaces: Tuple[str, ...],
        accept_partial: bool
    ) -> Tuple[Any, List[LookupTrace]]:

    inner_hint = extract_inner_hint(hint)

    if explicit_value is not None:
        value = explicit_value
        traces: List[LookupTrace] = []
    else:
        # resolve key value via active providers passing the original hint ie. to preserve TSecretValue
        value, traces = _resolve_single_value(key, hint, inner_hint, config_namespace, explicit_namespaces, embedded_namespaces)
        log_traces(config, key, hint, value, default_value, traces)
    # contexts must be resolved as a whole
    if is_context_inner_hint(inner_hint):
        pass
    # if inner_hint is BaseConfiguration then resolve it recursively
    elif is_base_configuration_inner_hint(inner_hint):
        if isinstance(value, BaseConfiguration):
            # if resolved value is instance of configuration (typically returned by context provider)
            embedded_config = value
            value = None
        elif isinstance(default_value, BaseConfiguration):
            # if default value was instance of configuration, use it
            embedded_config = default_value
            default_value = None
        else:
            embedded_config = inner_hint()

        if embedded_config.is_resolved():
            # injected context will be resolved
            value = embedded_config
        else:
            # only config with namespaces may look for initial values
            if embedded_config.__namespace__ and value is None:
                # config namespace becomes the key if the key does not start with, otherwise it keeps its original value
                initial_key, initial_embedded = _apply_embedded_namespaces_to_config_namespace(embedded_config.__namespace__, embedded_namespaces + (key,))
                # it must be a secret value is config is credentials
                initial_hint = TSecretValue if isinstance(embedded_config, CredentialsConfiguration) else AnyType
                value, initial_traces = _resolve_single_value(initial_key, initial_hint, AnyType, None, explicit_namespaces, initial_embedded)
                if isinstance(value, C_Mapping):
                    # mappings are not passed as initials
                    value = None
                else:
                    traces.extend(initial_traces)
                    log_traces(config, initial_key, type(embedded_config), value, default_value, initial_traces)

            # check if hint optional
            is_optional = is_optional_type(hint)
            # accept partial becomes True if type if optional so we do not fail on optional configs that do not resolve fully
            accept_partial = accept_partial or is_optional
            # create new instance and pass value from the provider as initial, add key to namespaces
            value = _resolve_configuration(embedded_config, explicit_namespaces, embedded_namespaces + (key,), default_value if value is None else value, accept_partial)
    else:
        # if value is resolved, then deserialize and coerce it
        if value is not None:
            # do not deserialize explicit values
            if value is not explicit_value:
                value = deserialize_value(key, value, inner_hint)

    return default_value if value is None else value, traces


def _call_method_in_mro(config: BaseConfiguration, method_name: str) -> None:
    # python multi-inheritance is cooperative and this would require that all configurations cooperatively
    # call each other class_method_name. this is not at all possible as we do not know which configs in the end will
    # be mixed together.

    # get base classes in order of derivation
    mro = type.mro(type(config))
    for c in mro:
        # check if this class implements on_resolved (skip pure inheritance to not do double work)
        if method_name in c.__dict__ and callable(getattr(c, method_name)):
            # pass right class instance
            c.__dict__[method_name](config)


def _resolve_single_value(
        key: str,
        hint: Type[Any],
        inner_hint: Type[Any],
        config_namespace: str,
        explicit_namespaces: Tuple[str, ...],
        embedded_namespaces: Tuple[str, ...]
    ) -> Tuple[Optional[Any], List[LookupTrace]]:

    traces: List[LookupTrace] = []
    value = None

    container = Container()
    # get providers from container
    providers_context = container[ConfigProvidersContext]
    # we may be resolving context
    if is_context_inner_hint(inner_hint):
        # resolve context with context provider and do not look further
        value, _ = providers_context.context_provider.get_value(key, inner_hint)
        return value, traces
    if is_base_configuration_inner_hint(inner_hint):
        # cannot resolve configurations directly
        return value, traces

    # resolve a field of the config
    config_namespace, embedded_namespaces = _apply_embedded_namespaces_to_config_namespace(config_namespace, embedded_namespaces)
    providers = providers_context.providers
    # get additional namespaces to look in from container
    namespaces_context = container[ConfigNamespacesContext]

    def look_namespaces(pipeline_name: str = None) -> Any:
        # start looking from the top provider with most specific set of namespaces first
        for provider in providers:
            value, provider_traces = resolve_single_provider_value(
                provider,
                key,
                hint,
                pipeline_name,
                config_namespace,
                # if explicit namespaces are provided, ignore the injected context
                explicit_namespaces or namespaces_context.namespaces,
                embedded_namespaces
            )
            traces.extend(provider_traces)
            if value is not None:
                # value found, ignore other providers
                break

        return value

    # first try with pipeline name as namespace, if present
    if namespaces_context.pipeline_name:
        value = look_namespaces(namespaces_context.pipeline_name)
    # then without it
    if value is None:
        value = look_namespaces()

    return value, traces


def resolve_single_provider_value(
    provider: ConfigProvider,
    key: str,
    hint: Type[Any],
    pipeline_name: str = None,
    config_namespace: str = None,
    explicit_namespaces: Tuple[str, ...] = (),
    embedded_namespaces: Tuple[str, ...] = (),
    # context_namespaces: Tuple[str, ...] = (),
    ) -> Tuple[Optional[Any], List[LookupTrace]]:
    traces: List[LookupTrace] = []

    if provider.supports_namespaces:
        ns = list(explicit_namespaces)
        # always extend with embedded namespaces
        ns.extend(embedded_namespaces)
    else:
        # if provider does not support namespaces and pipeline name is set then ignore it
        if pipeline_name:
            return None, traces
        else:
            # pass empty namespaces
            ns = []

    value = None
    while True:
        if (pipeline_name or config_namespace) and provider.supports_namespaces:
            full_ns = ns.copy()
            # pipeline, when provided, is the most outer and always present
            if pipeline_name:
                full_ns.insert(0, pipeline_name)
            # config namespace, is always present and innermost
            if config_namespace:
                full_ns.append(config_namespace)
        else:
            full_ns = ns
        value, ns_key = provider.get_value(key, hint, *full_ns)
        # if secret is obtained from non secret provider, we must fail
        cant_hold_it: bool = not provider.supports_secrets and is_secret_hint(hint)
        if value is not None and cant_hold_it:
            raise ValueNotSecretException(provider.name, ns_key)

        # create trace, ignore providers that cant_hold_it
        if not cant_hold_it:
            traces.append(LookupTrace(provider.name, full_ns, ns_key, value))

        if value is not None:
            # value found, ignore further namespaces
            break
        if len(ns) == 0:
            # namespaces exhausted
            break
        # pop optional namespaces for less precise lookup
        ns.pop()

    return value, traces


def _apply_embedded_namespaces_to_config_namespace(config_namespace: str, embedded_namespaces: Tuple[str, ...]) -> Tuple[str, Tuple[str, ...]]:
    # for the configurations that have __namespace__ (config_namespace) defined and are embedded in other configurations,
    # the innermost embedded namespace replaces config_namespace
    if embedded_namespaces:
        # do not add key to embedded namespaces if it starts with _, those namespaces must be ignored
        if not embedded_namespaces[-1].startswith("_"):
            config_namespace = embedded_namespaces[-1]
        embedded_namespaces = embedded_namespaces[:-1]

    # remove all embedded ns starting with _
    return config_namespace, tuple(ns for ns in embedded_namespaces if not ns.startswith("_"))



from typing import Dict

from dlt.common.schema.schema import Schema
from dlt.common.storages.schema_storage import SchemaStorage
from dlt.common.configuration.specs import SchemaVolumeConfiguration
from dlt.common.configuration.accessors import config


class LiveSchemaStorage(SchemaStorage):

    def __init__(self, config: SchemaVolumeConfiguration = config.value, makedirs: bool = False) -> None:
        self.live_schemas: Dict[str, Schema] = {}
        super().__init__(config, makedirs)

    def __getitem__(self, name: str) -> Schema:
        if name in self.live_schemas:
            schema = self.live_schemas[name]
        else:
            # return new schema instance
            schema = super().load_schema(name)
            self._update_live_schema(schema)

        return schema

    def load_schema(self, name: str) -> Schema:
        self.commit_live_schema(name)
        # now live schema is saved so we can load it with the changes
        return super().load_schema(name)

    def save_schema(self, schema: Schema) -> str:
        rv = super().save_schema(schema)
        # update the live schema with schema being saved, if no live schema exist, create one to be available for a getter
        self._update_live_schema(schema)
        return rv

    def remove_schema(self, name: str) -> None:
        super().remove_schema(name)
        # also remove the live schema
        self.live_schemas.pop(name, None)

    def initialize_import_schema(self, schema: Schema) -> None:
        if self.config.import_schema_path:
            try:
                self._load_import_schema(schema.name)
            except FileNotFoundError:
                # save import schema only if it not exist
                self._export_schema(schema, self.config.import_schema_path)

    def commit_live_schema(self, name: str) -> Schema:
        # if live schema exists and is modified then it must be used as an import schema
        live_schema = self.live_schemas.get(name)
        if live_schema and live_schema.stored_version_hash != live_schema.version_hash:
            live_schema.bump_version()
            self._save_schema(live_schema)
        return live_schema

    def _update_live_schema(self, schema: Schema, can_create_new: bool = True) -> None:
        live_schema = self.live_schemas.get(schema.name)
        if live_schema:
            # replace content without replacing instance
            # print(f"live schema {live_schema} updated in place")
            live_schema.replace_schema_content(schema)
        elif can_create_new:
            # print(f"live schema {schema.name} created from schema")
            self.live_schemas[schema.name] = schema



from inspect import Signature
from typing import Any, Set, Type
from dlt.common.exceptions import DltException


class ExtractorException(DltException):
    pass


class DltSourceException(DltException):
    pass


class DltResourceException(DltSourceException):
    def __init__(self, resource_name: str, msg: str) -> None:
        self.resource_name = resource_name
        super().__init__(msg)


class PipeException(DltException):
    def __init__(self, pipe_name: str, msg: str) -> None:
        self.pipe_name = pipe_name
        msg = f"In processing pipe {pipe_name}: " + msg
        super().__init__(msg)


class CreatePipeException(PipeException):
    def __init__(self, pipe_name: str, msg: str) -> None:
        super().__init__(pipe_name, msg)


class PipeItemProcessingError(PipeException):
    def __init__(self, pipe_name: str, msg: str) -> None:
        super().__init__(pipe_name, msg)


class PipeNotBoundToData(PipeException):
    def __init__(self, pipe_name: str, has_parent: bool) -> None:
        self.pipe_name = pipe_name
        self.has_parent = has_parent
        if has_parent:
            msg = f"A pipe created from transformer {pipe_name} is unbound or its parent is unbound or empty. Provide a resource in `data_from` argument or bind resources with | operator."
        else:
            msg = "Pipe is empty and does not have a resource at its head"
        super().__init__(pipe_name, msg)


class InvalidStepFunctionArguments(PipeException):
    def __init__(self, pipe_name: str, func_name: str, sig: Signature, call_error: str) -> None:
        self.func_name = func_name
        self.sig = sig
        super().__init__(pipe_name, f"Unable to call {func_name}: {call_error}. The mapping/filtering function {func_name} requires first argument to take data item and optional second argument named 'meta', but the signature is {sig}")


class ResourceNameMissing(DltResourceException):
    def __init__(self) -> None:
        super().__init__(None, """Resource name is missing. If you create a resource directly from data ie. from a list you must pass the name explicitly in `name` argument.
        Please note that for resources created from functions or generators, the name is the function name by default.""")


# class DependentResourceIsNotCallable(DltResourceException):
#     def __init__(self, resource_name: str) -> None:
#         super().__init__(resource_name, f"Attempted to call the dependent resource {resource_name}. Do not call the dependent resources. They will be called only when iterated.")


class ResourceNotFoundError(DltResourceException, KeyError):
      def __init__(self, resource_name: str, context: str) -> None:
          self.resource_name = resource_name
          super().__init__(resource_name, f"Resource with a name {resource_name} could not be found. {context}")


class InvalidResourceDataType(DltResourceException):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any], msg: str) -> None:
        self.item = item
        self._typ = _typ
        super().__init__(resource_name, f"Cannot create resource {resource_name} from specified data. " + msg)


class InvalidResourceDataTypeAsync(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any,_typ: Type[Any]) -> None:
        super().__init__(resource_name, item, _typ, "Async iterators and generators are not valid resources. Please use standard iterators and generators that yield Awaitables instead (for example by yielding from async function without await")


class InvalidResourceDataTypeBasic(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any,_typ: Type[Any]) -> None:
        super().__init__(resource_name, item, _typ, f"Resources cannot be strings or dictionaries but {_typ.__name__} was provided. Please pass your data in a list or as a function yielding items. If you want to process just one data item, enclose it in a list.")


class InvalidResourceDataTypeFunctionNotAGenerator(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any,_typ: Type[Any]) -> None:
        super().__init__(resource_name, item, _typ, "Please make sure that function decorated with @dlt.resource uses 'yield' to return the data.")


class InvalidResourceDataTypeMultiplePipes(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any,_typ: Type[Any]) -> None:
        super().__init__(resource_name, item, _typ, "Resources with multiple parallel data pipes are not yet supported. This problem most often happens when you are creating a source with @dlt.source decorator that has several resources with the same name.")


class InvalidTransformerDataTypeGeneratorFunctionRequired(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any,_typ: Type[Any]) -> None:
        super().__init__(resource_name, item, _typ,
            "Transformer must be a function decorated with @dlt.transformer that takes data item as its first argument. Only first argument may be 'positional only'.")


class InvalidTransformerGeneratorFunction(DltResourceException):
    def __init__(self, resource_name: str, func_name: str, sig: Signature, code: int) -> None:
        self.func_name = func_name
        self.sig = sig
        self.code = code
        msg = f"Transformer function {func_name} must take data item as its first argument. "
        if code == 1:
            msg += "The actual function does not take any arguments."
        elif code == 2:
            msg += f"Only the first argument may be 'positional only', actual signature is {sig}"
        elif code == 3:
            msg += f"The first argument cannot be keyword only, actual signature is {sig}"

        super().__init__(resource_name, msg)


class InvalidResourceDataTypeIsNone(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(resource_name, item, _typ, "Resource data missing. Did you forget the return statement in @dlt.resource decorated function?")


class ResourceFunctionExpected(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(resource_name, item, _typ, f"Expected function or callable as first parameter to resource {resource_name} but {_typ.__name__} found. Please decorate a function with @dlt.resource")


class InvalidParentResourceDataType(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any,_typ: Type[Any]) -> None:
        super().__init__(resource_name, item, _typ, f"A parent resource of {resource_name} is of type {_typ.__name__}. Did you forget to use '@dlt.resource` decorator or `resource` function?")


class InvalidParentResourceIsAFunction(DltResourceException):
    def __init__(self, resource_name: str, func_name: str) -> None:
        self.func_name = func_name
        super().__init__(resource_name, f"A data source {func_name} of a transformer {resource_name} is an undecorated function. Please decorate it with '@dlt.resource' or pass to 'resource' function.")


class DeletingResourcesNotSupported(DltResourceException):
    def __init__(self, source_name: str, resource_name: str) -> None:
        super().__init__(resource_name, f"Resource cannot be removed the the source {source_name}")


class ParametrizedResourceUnbound(DltResourceException):
    def __init__(self, resource_name: str, func_name: str, sig: Signature, kind: str = "resource") -> None:
        self.func_name = func_name
        self.sig = sig
        super().__init__(resource_name, f"The {kind} {resource_name} is parametrized and expects following arguments: {sig}. Did you forget to call the {func_name} function? For example from `source.{resource_name}(...)")


class TableNameMissing(DltSourceException):
    def __init__(self) -> None:
        super().__init__("""Table name is missing in table template. Please provide a string or a function that takes a data item as an argument""")


class InconsistentTableTemplate(DltSourceException):
    def __init__(self, reason: str) -> None:
        msg = f"A set of table hints provided to the resource is inconsistent: {reason}"
        super().__init__(msg)


class DataItemRequiredForDynamicTableHints(DltResourceException):
    def __init__(self, resource_name: str) -> None:
        super().__init__(resource_name, f"""An instance of resource's data required to generate table schema in resource {resource_name}.
        One of table hints for that resource (typically table name) is a function and hint is computed separately for each instance of data extracted from that resource.""")


class SourceDataIsNone(DltSourceException):
    def __init__(self, source_name: str) -> None:
        self.source_name = source_name
        super().__init__(f"No data returned or yielded from source function {source_name}. Did you forget the return statement?")


class SourceExhausted(DltSourceException):
    def __init__(self, source_name: str) -> None:
        self.source_name = source_name
        super().__init__(f"Source {source_name} is exhausted or has active iterator. You can iterate or pass the source to dlt pipeline only once.")


class ResourcesNotFoundError(DltSourceException):
    def __init__(self, source_name: str, available_resources: Set[str], requested_resources: Set[str]) -> None:
        self.source_name = source_name
        self.available_resources = available_resources
        self.requested_resources = requested_resources
        self.not_found_resources = requested_resources.difference(available_resources)
        msg = f"The following resources could not be found in source {source_name}: {self.not_found_resources}. Available resources are: {available_resources}"
        super().__init__(msg)


class SourceNotAFunction(DltSourceException):
    def __init__(self, source_name: str, item: Any, _typ: Type[Any]) -> None:
        self.source_name = source_name
        self.item = item
        self.typ = _typ
        super().__init__(f"First parameter to the source {source_name} must be a function or callable but is {_typ.__name__}. Please decorate a function with @dlt.source")


class SourceIsAClassTypeError(DltSourceException):
    def __init__(self, source_name: str,  _typ: Type[Any]) -> None:
        self.source_name = source_name
        self.typ = _typ
        super().__init__(f"First parameter to the source {source_name} is a class {_typ.__name__}. Do not decorate classes with @dlt.source. Instead implement __call__ in your class and pass instance of such class to dlt.source() directly")
